class BaseAttachment(object):
    def __init__(self, attachment_type, url):
        self.type = attachment_type
        self.url = url

    def to_dict(self):
        return {
            'attachment': {
                'type': self.type,
                'payload': {
                    'url': self.url
                }
            }
        }


class Image(BaseAttachment):
    def __init__(self, url):
        self.attachment_type = 'image'
        self.url = url
        super(Image, self).__init__(self.attachment_type, self.url)


class Audio(BaseAttachment):
    def __init__(self, url):
        self.attachment_type = 'audio'
        self.url = url
        super(Audio, self).__init__(self.attachment_type, self.url)


class Video(BaseAttachment):
    def __init__(self, url):
        self.attachment_type = 'video'
        self.url = url
        super(Video, self).__init__(self.attachment_type, self.url)


class File(BaseAttachment):
    def __init__(self, url):
        self.attachment_type = 'file'
        self.url = url
        super(File, self).__init__(self.attachment_type, self.url)



from __future__ import absolute_import
import abc
import logging

import requests

__version__ = '2.0.0'

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class MessengerClient(object):

    def __init__(self, page_access_token):
        self.page_access_token = page_access_token

    def get_user_data(self, entry):
        r = requests.get(
            'https://graph.facebook.com/v2.6/{sender}'.format(sender=entry['sender']['id']),
            params={
                'fields': 'first_name,last_name,profile_pic,locale,timezone,gender',
                'access_token': self.page_access_token
            }
        )
        return r.json()

    def send(self, payload, entry):
        r = requests.post(
            'https://graph.facebook.com/v2.6/me/messages',
            params={
                'access_token': self.page_access_token
            },
            json={
                'recipient': {
                    'id': entry['sender']['id']
                },
                'message': payload
            }
        )
        return r.json()

    def send_action(self, sender_action, entry):
        r = requests.post(
            'https://graph.facebook.com/v2.6/me/messages',
            params={
                'access_token': self.page_access_token
            },
            json={
                'recipient': {
                    'id': entry['sender']['id']
                },
                'sender_action': sender_action
            }
        )
        return r.json()

    def subscribe_app_to_page(self):
        r = requests.post(
            'https://graph.facebook.com/v2.6/me/subscribed_apps',
            params={
                'access_token': self.page_access_token
            }
        )
        return r.json()

    def set_thread_setting(self, data):
        r = requests.post(
            'https://graph.facebook.com/v2.6/me/thread_settings',
            params={
                'access_token': self.page_access_token
            },
            json=data
        )
        return r.json()

    def delete_get_started(self):  # pragma: no cover
        r = requests.delete(
            'https://graph.facebook.com/v2.6/me/thread_settings',
            params={
                'access_token': self.page_access_token
            },
            json={
                'setting_type': 'call_to_actions',
                'thread_state': 'new_thread'
            }
        )
        return r.json()

    def link_account(self, account_linking_token):
        r = requests.post(
            'https://graph.facebook.com/v2.6/me',
            params={
                'access_token': self.page_access_token,
                'fields': 'recipient',
                'account_linking_token': account_linking_token
            }
        )
        return r.json()

    def unlink_account(self, psid):
        r = requests.post(
            'https://graph.facebook.com/v2.6/me/unlink_accounts',
            params={
                'access_token': self.page_access_token
            },
            json={
                'psid': psid
            }
        )
        return r.json()


class BaseMessenger(object):
    __metaclass__ = abc.ABCMeta

    last_message = {}

    def __init__(self, page_access_token):
        self.page_access_token = page_access_token
        self.client = MessengerClient(self.page_access_token)

    @abc.abstractmethod
    def account_linking(self, message):
        """Method to handle `account_linking`"""

    @abc.abstractmethod
    def messages(self, message):
        """Method to handle `messages`"""

    @abc.abstractmethod
    def message_deliveries(self, message):
        """Method to handle `message_deliveries`"""

    @abc.abstractmethod
    def messaging_optins(self, message):
        """Method to handle `messaging_optins`"""

    @abc.abstractmethod
    def messaging_postbacks(self, message):
        """Method to handle `messaging_postbacks`"""

    @abc.abstractmethod
    def message_reads(self, message):
        """Method to handle `message_reads`"""

    def handle(self, payload):
        for entry in payload['entry']:
            for message in entry['messaging']:
                self.last_message = message
                if message.get('account_linking'):
                    return self.account_linking(message)
                elif message.get('delivery'):
                    return self.message_deliveries(message)
                elif message.get('message'):
                    return self.messages(message)
                elif message.get('optin'):
                    return self.messaging_optins(message)
                elif message.get('postback'):
                    return self.messaging_postbacks(message)
                elif message.get('read'):
                    return self.message_reads(message)

    def get_user(self):
        return self.client.get_user_data(self.last_message)

    def send(self, payload):
        return self.client.send(payload, self.last_message)

    def send_action(self, sender_action):
        return self.client.send_action(sender_action, self.last_message)

    def get_user_id(self):
        return self.last_message['sender']['id']

    def subscribe(self):
        return self.client.subscribe_app_to_page()

    def set_thread_setting(self, data):
        return self.client.set_thread_setting(data)

    def delete_get_started(self):
        return self.client.delete_get_started()

    def link_account(self, account_linking_token):
        return self.client.link_account(account_linking_token)

    def unlink_account(self, psid):
        return self.client.unlink_account(psid)



class Text(object):
    def __init__(self, text):
        self.text = text

    def to_dict(self):
        return {
            'text': self.text
        }


class Button(object):
    BUTTON_TYPES = [
        'web_url',
        'postback',
        'phone_number',
        'account_link',
        'account_unlink'
    ]

    def __init__(self, button_type, title, url=None, payload=None):
        if button_type not in self.BUTTON_TYPES:
            raise ValueError('Invalid button_type provided.')
        if len(title) > 20:
            raise ValueError('Title cannot be longer 20 characters.')
        self.type = button_type
        self.title = title
        self.url = url
        self.payload = payload

    def to_dict(self):
        res = {
            'type': self.type,
            'title': self.title
        }
        if self.url:
            res['url'] = self.url
        if self.payload:
            res['payload'] = self.payload
        return res


class Element(object):
    """
    To be used with the generic template to create a carousel
    """

    def __init__(self, title, item_url=None, image_url=None,
                 subtitle=None, buttons=None, quantity=None,
                 price=None, currency=None):

        self._title = title
        self.item_url = item_url
        self.image_url = image_url
        self._subtitle = subtitle
        self.buttons = buttons
        self.quantity = quantity
        self.price = price
        self.currency = currency

    @property
    def title(self):
        if len(self._title) > 45:
            raise ValueError(
                'Title cannot be longer 45 characters'
            )
        return self._title

    @property
    def subtitle(self):
        if self._subtitle:
            if len(self._subtitle) > 80:
                raise ValueError('Subtitle cannot be longer 80 characters')
        return self._subtitle

    def to_dict(self):
        data = {
            'title': self.title,
        }
        if self.item_url:
            data['item_url'] = self.item_url
        if self.image_url:
            data['image_url'] = self.image_url
        if self._subtitle:
            data['subtitle'] = self.subtitle
        if self.quantity:
            data['quantity'] = self.quantity
        if self.price:
            data['price'] = self.price
        if self.currency:
            data['currency'] = self.currency
        if self.buttons:
            data['buttons'] = [
                button.to_dict() for button in self.buttons
            ]

        return data


class Adjustment(object):
    def __init__(self, name=None, amount=None):
        # Optional
        self.name = name
        self.amount = amount

    def to_dict(self):
        return {
            'name': self.name,
            'amount': self.amount
        }


class Address(object):
    def __init__(self, street_1, city, postal_code,
                 state, country, street_2=''):
        # Required
        self.street_1 = street_1
        self.city = city
        self.postal_code = postal_code
        self.state = state
        self.country = country
        # Optional
        self.street_2 = street_2

    def to_dict(self):
        return {
            'street_1': self.street_1,
            'street_2': self.street_2,
            'city': self.city,
            'postal_code': self.postal_code,
            'state': self.state,
            'country': self.country,
        }


class Summary(object):
    def __init__(self, total_cost, subtotal=None, shipping_cost=None, total_tax=None):
        # Required
        self.total_cost = total_cost
        # Optional
        self.subtotal = subtotal
        self.shipping_cost = shipping_cost
        self.total_tax = total_tax

    def to_dict(self):
        return {
            'subtotal': self.subtotal,
            'shipping_cost': self.shipping_cost,
            'total_tax': self.total_tax,
            'total_cost': self.total_cost
        }



class GreetingText(object):
    def __init__(self, text):
        if len(text) > 160:
            raise ValueError('Text cannot be longer 160 characters.')
        self.text = text

    def to_dict(self):
        return {
            'setting_type': 'greeting',
            'greeting': {
                'text': self.text
            }
        }


class GetStartedButton(object):
    def __init__(self, payload):
        self.payload = payload

    def to_dict(self):
        return {
            'setting_type': 'call_to_actions',
            'thread_state': 'new_thread',
            'call_to_actions': [
                {
                    'payload': self.payload
                }
            ]
        }


class PersistentMenuItem(object):
    ITEM_TYPES = [
        'web_url',
        'postback'
    ]

    def __init__(self, item_type, title, url=None, payload=None):
        if item_type not in self.ITEM_TYPES:
            raise ValueError('Invalid item_type provided.')
        if len(title) > 30:
            raise ValueError('Title cannot be longer 30 characters.')
        if payload and len(payload) > 1000:
            raise ValueError('Payload cannot be longer 1000 characters.')
        if item_type == 'web_url' and url is None:
            raise ValueError('`url` must be supplied for `web_url` type menu items.')
        if item_type == 'postback' and payload is None:
            raise ValueError('`postback` must be supplied for `payload` type menu items.')

        self.item_type = item_type
        self.title = title
        self.url = url
        self.payload = payload

    def to_dict(self):
        res = {
            'type': self.item_type,
            'title': self.title
        }

        if self.url and self.item_type == 'web_url':
            res['url'] = self.url

        if self.payload and self.item_type == 'postback':
            res['payload'] = self.payload
        return res


class PersistentMenu(object):
    def __init__(self, menu_items):
        if len(menu_items) > 5:
            raise ValueError('You cannot have more than 5 menu_items.')
        self.menu_items = menu_items

    def to_dict(self):
        return {
            'setting_type': 'call_to_actions',
            'thread_state': 'existing_thread',
            'call_to_actions': [
                item.to_dict() for item in self.menu_items
            ]
        }



class BaseTemplate(object):
    def __init__(self, elements):
        if not isinstance(elements, list):
            elements = [elements]
        self._elements = elements

    @property
    def elements(self):
        if len(self._elements) > 10:
            raise ValueError('You cannot have more than 10 elements in the template.')
        return self._elements


class GenericTemplate(BaseTemplate):

    def __init__(self, elements):
        self._elements = elements
        super(GenericTemplate, self).__init__(self._elements)

    def to_dict(self):
        return {
            'attachment': {
                'type': 'template',
                'payload': {
                    'template_type': 'generic',
                    'elements': [
                        element.to_dict() for element in self.elements
                    ]
                }
            },
        }


class ButtonTemplate(object):

    def __init__(self, text, buttons):
        self.text = text

        if not isinstance(buttons, list):
            buttons = [buttons]
        self.buttons = buttons

    def to_dict(self):
        return {
            'attachment': {
                'type': 'template',
                'payload': {
                    'template_type': 'button',
                    'text': self.text,
                    'buttons': [
                        button.to_dict() for button in self.buttons
                    ]
                }
            }
        }


class ReceiptTemplate(BaseTemplate):

    def __init__(self, recipient_name, order_number, currency, payment_method,
                 elements, summary, order_url=None, timestamp=None,
                 address=None, adjustments=None):

        self._elements = elements
        self.recipient_name = recipient_name
        self.order_number = order_number
        self.currency = currency
        self.payment_method = payment_method
        self.summary = summary.to_dict()
        self.order_url = order_url
        self.timestamp = timestamp
        self.address = {} if address is None else address.to_dict()
        self.adjustments = [] if adjustments is None else adjustments

        super(ReceiptTemplate, self).__init__(self._elements)

    def to_dict(self):
        return {
            'attachment': {
                'type': 'template',
                'payload': {
                    'template_type': 'receipt',
                    'recipient_name': self.recipient_name,
                    'order_number': self.order_number,
                    'order_url': self.order_url,
                    'currency': self.currency,
                    'timestamp': self.timestamp,
                    'payment_method': self.payment_method,
                    'elements': [
                        element.to_dict() for element in self.elements
                    ],
                    'adjustments': [
                        adjustment.to_dict() for adjustment in self.adjustments
                    ],
                    'address': self.address,
                    'summary': self.summary
                }
            },
        }



class QuickReply(object):

    def __init__(self, title, payload):
        if len(title) > 20:
            raise ValueError('Title cannot be longer 20 characters.')
        if len(payload) > 1000:
            raise ValueError('Payload cannot be longer 1000 characters.')

        self.title = title
        self.payload = payload

    def to_dict(self):
        return {
            'content_type': 'text',
            'title': self.title,
            'payload': self.payload
        }


class QuickReplies(object):
    def __init__(self, quick_replies):
        if len(quick_replies) > 10:
            raise ValueError('You cannot have more than 10 quick replies.')
        self.quick_replies = quick_replies

    def to_dict(self):
        return {
            'quick_replies': [
                quick_reply.to_dict() for quick_reply in self.quick_replies
            ]
        }
